from distutils.log import warn
import warnings
import ast

try:
    # Useful for debugging
    import astor
except ImportError:
    pass

from collections import defaultdict
from typing import Callable, Dict, List, Optional, Sequence, Set
from abc import ABC, abstractmethod
from ffcv.pipeline.allocation_query import AllocationQuery

from ffcv.pipeline.pipeline_spec import PipelineSpec
from ffcv.pipeline.compiler import Compiler
from ffcv.pipeline.allocation_query import allocate_query
from .operation import Operation
from ..transforms import ModuleWrapper
from .state import State

import torch as ch
import numpy as np

# This is the starting state of the pipeline
INITIAL_STATE = State(jit_mode=True,
                       device=ch.device('cpu'),
                       dtype=np.dtype('u1'),
                       shape=None)


class Node(ABC):
    last_node_id: int = 0
    def __init__(self):
        self.id = Node.last_node_id
        self._code = None
        Node.last_node_id += 1

    @property
    @abstractmethod
    def is_jitted(self):
        raise NotImplemented()

    @property
    @abstractmethod
    def parent(self):
        raise NotImplemented()
    
    @property
    @abstractmethod
    def arg_id(self):
        raise NotImplemented()
    
    @property
    @abstractmethod
    def result_id(self):
        raise NotImplemented()
    
    @property
    @abstractmethod
    def result_id(self):
        raise NotImplemented()
    
    def get_shared_code_ast(self, done_ops):
        return ast.Pass()
    
    @abstractmethod
    def generate_code(self):
        raise NotImplemented()

    def recompile(self):
        self._code = self.generate_code()

    @property
    def with_indices(self):
        try:
            return self.code.with_indices
        except:
            return False

    @property
    def code(self):
        if self._code is None:
            self.recompile()

        return self._code

    @property
    def func_call_ast(self):
        pipeline_identifier = f'code_{self.id}'
        memory_identifier = f'memory_{self.id}'

        tree = ast.parse(f"""
{self.result_id} = {pipeline_identifier}({self.arg_id}, {memory_identifier})
        """).body[0]

        if self.with_indices:
            tree.value.args.extend([
                ast.Name(id='batch_indices', ctx=ast.Load()),
            ])
        return tree


class DecoderNode(Node):
    def __init__(self, field_name:str, decoder: Operation, f_ix:int):
        super().__init__()
        self.field_name = field_name
        self.decoder = decoder
        self.f_ix = f_ix

    @property
    def is_jitted(self):
        # Decoder have to jitted
        return True

    @property
    def parent(self):
        return None

    @property
    def arg_id(self):
        return 'batch_indices'

    @property
    def result_id(self):
        return f"result_{self.id}"

    def generate_code(self):
        return self.decoder.generate_code()

    @property
    def func_call_ast(self):
        tree = super().func_call_ast
        tree.value.args.extend([
            ast.Subscript(value=ast.Name(id='metadata', ctx=ast.Load()),
                          slice=ast.Index(value=ast.Constant(value=f'f{self.f_ix}', kind=None)), ctx=ast.Load()),
                 ast.Name(id='storage_state', ctx=ast.Load()),
        ])

        return tree


class TransformNode(Node):
    def __init__(self, parent:Node, operation: Operation):
        super().__init__()
        self._parent = parent
        self.operation = operation
        self.jitted = True

    def __repr__(self):
        return f'TransformerNode({self.operation})'

    def generate_code(self):
        return self.operation.generate_code()

    @property
    def parent(self):
        return self._parent

    @property
    def is_jitted(self):
        # Decoder have to jitted
        return self.jitted

    @property
    def arg_id(self):
        return self.parent.result_id

    @property
    def result_id(self):
        return f"result_{self.id}"

    def get_shared_code_ast(self, done_ops):
        if self.operation in done_ops:
            return ast.Pass()

        done_ops[self.operation] = self.id

        pipeline_identifier = f'init_shared_state_code_{self.id}'
        memory_identifier = f'shared_memory_{self.id}'

        tree = ast.parse(f"""{pipeline_identifier}({memory_identifier})""").body[0]


        return tree


class RefNode(Node):
    def __init__(self, ref_operation: Operation):
        super().__init__()
        self.ref_operation = ref_operation
        self._parent = None

    def resolve_operation(self, operation_to_node: Dict[Operation, List[Node]]):
        entries = operation_to_node[self.ref_operation]
        if not entries:
            raise ValueError(f"{self.ref_operation} not found in other pipelines")
        if len(entries) > 1:
            raise ValueError(f"Reference to {self.ref_operation} ambiguous")

        self._parent = entries[0]

    @property
    def parent(self):
        assert self._parent is not None
        return self._parent

    @property
    def is_jitted(self):
        # RefNodes can be either jitted or not,
        # whatever produces smaller pipelines
        return None

    @property
    def arg_id(self):
        return None  # Ref's don't have arguments

    def generate_code(self):
        def nop(*args, **kwargs):
            return None

    @property
    def func_call_ast(self):
        return ast.Pass()

    @property
    def result_id(self):
        return self.parent.result_id


class Graph:

    def __init__(self, pipeline_specs: Dict[str, PipelineSpec], handlers,
                 fieldname_to_fix, metadata, memory_read):

        self.memory_read = memory_read
        self.handlers = handlers
        self.fieldname_to_fix = fieldname_to_fix
        self.metadata = metadata
        self.pipeline_specs = pipeline_specs
        self.nodes: List[Node] = []
        self.root_nodes: Dict[Node, str] = {}
        self.leaf_nodes: Dict[str, Node] = {}
        self.operation_to_node = defaultdict(list)
        self.id_to_node = {}
        self.node_to_id = {}

        # Filling the default decoders
        for output_name, spec in pipeline_specs.items():
            if spec.source in self.handlers:
                field = self.handlers[spec.source]
                Decoder = field.get_decoder_class()
                spec.accept_decoder(Decoder, output_name)

        # registering nodes
        for output_name, spec in pipeline_specs.items():
            if spec.source is None:
                raise ValueError(f"Field {output_name} has no source")

            source = spec.source
            # This pipeline starts with a decoder
            if isinstance(source, str):
                assert spec.decoder is not None
                node = DecoderNode(source, spec.decoder, fieldname_to_fix[source])
                self.operation_to_node[spec.decoder].append(node)
                self.root_nodes[node] = source
            else:
                node = RefNode(source)
                assert spec.decoder is None

            self.nodes.append(node)

            for operation in spec.transforms:
                node = TransformNode(node, operation)
                self.operation_to_node[operation].append(node)
                self.nodes.append(node)

            self.leaf_nodes[output_name] = node
            
        # resolve references
        for node in self.nodes:
            if isinstance(node, RefNode):
                node.resolve_operation(self.operation_to_node)

        # Filling the adjacency list
        self.adjacency_list = defaultdict(list)
        for node in self.nodes:
            self.id_to_node[node.id] = node
            self.node_to_id[node] = node.id
            if node.parent is not None:
                self.adjacency_list[node.parent].append(node)
                

    def collect_requirements(self, state=INITIAL_STATE,
                             current_node: Node = None,
                             allocations: Dict[int, Optional[AllocationQuery]] = None,
                             code: Dict[int, Optional[Callable]] = None,
                             source_field:str = None):

        if allocations is None:
            allocations: Dict[int, Optional[AllocationQuery]] = {
                'shared': {},
                'operation': {}
            }
        if code is None:
            code: Dict[int, Optional[Callable]] = {
                'shared': {},
                'operation': {}
            }
        next_state = state
        if current_node is None:
            next_nodes = self.root_nodes.keys()
        else:
            if not isinstance(current_node, RefNode):
                if isinstance(current_node, TransformNode):
                    operation = current_node.operation
                else:
                    operation = current_node.decoder

                if isinstance(current_node, DecoderNode):
                    source_field = current_node.field_name

                fix = self.fieldname_to_fix[source_field]
                metadata = self.metadata[f'f{fix}']

                operation.accept_field(self.handlers[source_field])
                operation.accept_globals(metadata, self.memory_read)

                next_state, allocation = operation.declare_state_and_memory(state)
                state_allocation = operation.declare_shared_memory(state)

                if next_state.device.type != 'cuda' and isinstance(operation,
                    ModuleWrapper):
                    msg = ("Using a pytorch transform on the CPU is extremely"
                        "detrimental to the performance, consider moving the augmentation"
                        "on the GPU or using an FFCV native transform")
                    warnings.warn(msg, ResourceWarning)


                if isinstance(current_node, TransformNode):
                    current_node.jitted = next_state.jit_mode

                allocations['operation'][current_node.id] = allocation
                allocations['shared'][current_node.id] = state_allocation
                code['operation'][current_node.id] = operation.generate_code()
                code['shared'][current_node.id] = operation.generate_code_for_shared_state()

            next_nodes = self.adjacency_list[current_node]

        for node in next_nodes:
            self.collect_requirements(next_state, node, allocations, code, source_field=source_field)

        return allocations, code

    def allocate_memory(self, batch_size, batches_ahead):

        memory_buffers = defaultdict(dict)
        full_memory_requirements, _ = self.collect_requirements()

        for kind, requirements in full_memory_requirements.items():
            for node_id, memory_allocation in requirements.items():
                # If the operation didn't make a query we stop here
                allocated_buffer = None
                if isinstance(memory_allocation, AllocationQuery):
                    allocated_buffer = allocate_query(memory_allocation,
                                                                batch_size,
                                                                batches_ahead)
                elif isinstance(memory_allocation, Sequence):
                    allocated_buffer = tuple(
                        allocate_query(q, batch_size, batches_ahead) for q in memory_allocation
                    )

                memory_buffers[kind][node_id] = allocated_buffer

        return memory_buffers

    def group_operations(self):
        current_front: Set[Node] = set()
        next_front: Set[Node] = set()
        stages = []

        for node in self.root_nodes.keys():
            current_front.add(node)


        while current_front:
            current_stage = list()
            jitted_stage = len(stages) % 2 == 0

            while current_front:
                node = current_front.pop()
                if node.is_jitted == jitted_stage or node.is_jitted is None:
                    current_stage.append(self.node_to_id[node])
                    current_front.update(set(self.adjacency_list[node]))

                else:
                    next_front.add(node)

            stages.append(current_stage)
            current_front = next_front

        return stages

    def codegen_stage(self, stage:List[Node], s_ix:int, op_to_node, code, already_defined):
        fun_name = f"stage_code_{s_ix}"
        base_code = ast.parse(f"""
def {fun_name}(batch_indices, metadata, storage_state):
    pass
        """).body[0]


        base_code.args.args.extend([
            ast.arg(arg=f'memory_{x}') for x in code['operation']
        ])

        base_code.args.args.extend([
            ast.arg(arg=f'shared_memory_{x}') for x in code['shared']
        ])

        base_code.args.args.extend([
            ast.arg(f'result_{x}') for x in already_defined
        ])

        return_tuple = ast.Return(value=ast.Tuple(elts=[], ctx=ast.Load()))

        defined_here = []

        base_code.body.pop()
        compiled_functions = {}
        for node_id in stage:
            node: Node = self.id_to_node[node_id]
            has_shared_state = node_id in code['shared'] and code['shared'][node_id] is not None

            try:
                compiled_functions[f'code_{node_id}'] = code['operation'][node_id]
            except KeyError:
                pass # No code for this node

            func_call_ast = node.func_call_ast
            if has_shared_state:
                fname = f'init_shared_state_code_{node_id}'
                compiled_functions[fname] = code['shared'][node_id]
                base_code.body.append(node.get_shared_code_ast(op_to_node))
                func_call_ast.value.args.extend([
                    ast.Name(id=f'shared_memory_{op_to_node[node.operation]}', ctx=ast.Load()),
                ])

            base_code.body.append(func_call_ast)
            return_tuple.value.elts.append(ast.Name(id=node.result_id, ctx=ast.Load()))
            already_defined.append(node.id)
            defined_here.append(node.id)

        # If the stage is even we are compiling it
        if s_ix % 2 == 0:
            compiled_functions = {k: Compiler.compile(v) for (k, v) in compiled_functions.items()}

        base_code.body.append(return_tuple)

        module = ast.fix_missing_locations(
            ast.Module(body=[base_code],
                       type_ignores=[])
        )

        # print(astor.to_source(base_code))
        namespace = {
            **compiled_functions
        }

        exec(compile(module, '', 'exec'), namespace)
        final_code = namespace[fun_name]
        return final_code, defined_here


    def codegen_all(self, code):
        stages = self.group_operations()
        code_stages = []
        already_defined = []

        # Set of operations that already had their state initialized
        # (We do not want to have their random state reset)
        op_to_node = {}

        for s_ix, stage in enumerate(stages):
            code_stages.append(self.codegen_stage(stage, s_ix, op_to_node, code, already_defined))

        final_output = [x.id for x in self.leaf_nodes.values()]
        return code_stages, final_output


import numpy as np
import torch as ch
from torch.utils.data import Dataset
from assertpy import assert_that
from tempfile import NamedTemporaryFile
from torchvision.datasets import CIFAR10
from torch.utils.data import Subset

from ffcv.writer import DatasetWriter
from ffcv.fields import IntField, RGBImageField
from ffcv.loader import Loader
from ffcv.pipeline.compiler import Compiler

class DummyDataset(Dataset):

    def __init__(self, length, height, width, reversed=False):
        self.length = length
        self.height = height
        self.width = width
        self.reversed = reversed
        
    def __len__(self):
        return self.length

    def __getitem__(self, index):
        if index > self.length:
            raise IndexError
        dims = (self.height, self.width, 3)
        image_data = ((np.ones(dims) * index) % 255).astype('uint8')
        result = index,image_data
        if self.reversed:
            result = tuple(reversed(result))
        return result

def create_and_validate(length, mode='raw', reversed=False):

    dataset = DummyDataset(length, 500, 300, reversed=reversed)

    with NamedTemporaryFile() as handle:
        name = handle.name
        
        fields = {
            'index': IntField(),
            'value': RGBImageField(write_mode=mode, jpeg_quality=95)
        }
        
        if reversed:
            fields = {
                'value': RGBImageField(write_mode=mode, jpeg_quality=95),
                'index': IntField()
            }

        writer = DatasetWriter(name, fields, num_workers=2)

        writer.from_indexed_dataset(dataset, chunksize=5)
            
        Compiler.set_enabled(False)
        
        loader = Loader(name, batch_size=5, num_workers=2)
        
        for res in loader:
            if not reversed:
                index, images  = res
            else:
                images , index = res

            for i, image in zip(index, images):
                if mode == 'raw':
                    assert_that(ch.all((image == (i % 255)).reshape(-1))).is_true()
                else:
                    assert_that(ch.all((image == (i % 255)).reshape(-1))).is_true()
                
def make_and_read_cifar_subset(length):
    my_dataset = Subset(CIFAR10(root='/tmp', train=True, download=True), range(length))

    with NamedTemporaryFile() as handle:
        name = handle.name
        writer = DatasetWriter(name, {
            'image': RGBImageField(write_mode='smart', 
                                max_resolution=32),
            'label': IntField(),
        }, num_workers=2)

        writer.from_indexed_dataset(my_dataset, chunksize=10)

        Compiler.set_enabled(False)
        
        loader = Loader(name, batch_size=5, num_workers=2)
        
        for index, images in loader:
            pass

def test_cifar_subset():
    make_and_read_cifar_subset(200)

def test_simple_raw_image_pipeline():
    create_and_validate(500, 'raw', False)

def test_simple_raw_image_pipeline_rev():
    create_and_validate(500, 'raw', True)

def test_simple_jpg_image_pipeline():
    create_and_validate(500, 'jpg', False)

def test_simple_jpg_image_pipeline_rev():
    create_and_validate(500, 'jpg', True)



"""
Example of defining a custom (image) transform using FFCV.
For tutorial, see https://docs.ffcv.io/ffcv_examples/custom_transforms.html.

"""
import time
import numpy as np
import torchvision

from ffcv.fields import IntField, RGBImageField
from ffcv.fields.decoders import SimpleRGBImageDecoder
from ffcv.loader import Loader, OrderOption
from ffcv.pipeline.compiler import Compiler
from ffcv.pipeline.operation import Operation, AllocationQuery
from ffcv.transforms import ToTensor
from ffcv.writer import DatasetWriter
from dataclasses import replace

class PickACorner(Operation):
    def generate_code(self):
        parallel_range = Compiler.get_iterator()
        def pick_a_corner(images, dst):
            which_corner = np.random.rand(images.shape[0])
            for i in parallel_range(images.shape[0]):
                if which_corner[i] == 0:
                    dst[i] = images[i,:images.shape[1]//2, :images.shape[2]//2]
                else:
                    dst[i] = images[i,-images.shape[1]//2:, -images.shape[2]//2:]

            return dst

        pick_a_corner.is_parallel = True
        return pick_a_corner

    def declare_state_and_memory(self, previous_state):
        h, w, c = previous_state.shape
        new_shape = (h // 2, w // 2, c)

        new_state = replace(previous_state, shape=new_shape)
        mem_allocation = AllocationQuery(new_shape, previous_state.dtype)
        return (new_state, mem_allocation)

# Step 1: Create an FFCV-compatible CIFAR-10 dataset
ds = torchvision.datasets.CIFAR10('/tmp', train=True, download=True)
writer = DatasetWriter('/tmp/cifar.beton', {
    'image': RGBImageField(),
    'label': IntField()
})
writer.from_indexed_dataset(ds)

# Step 2: Create data loaders
BATCH_SIZE = 512
# Create loaders
image_pipelines = {
    'with': [SimpleRGBImageDecoder(), PickACorner(), ToTensor()],
    'without': [SimpleRGBImageDecoder(), ToTensor()]
}

for name, pipeline in image_pipelines.items():
    loader = Loader(f'/tmp/cifar.beton', batch_size=BATCH_SIZE,
                    num_workers=16, order=OrderOption.RANDOM,
                    drop_last=True, pipelines={'image': pipeline})

    # First epoch includes compilation time
    for ims, labs in loader: pass
    start_time = time.time()
    for _ in range(100):
        for ims, labs in loader: pass
    print(f'Method: {name} | Shape: {ims.shape} | Time per epoch: {(time.time() - start_time) / 100:.5f}s')


from threading import Thread
from queue import Queue

import numpy as np

from ...libffcv import read


class PageReader(Thread):

    def __init__(self, fname:str, queries: Queue, loaded: Queue,
                 memory: np.ndarray):
        self.fname: str = fname
        self.queries: Queue = queries
        self.memory: np.ndarray = memory
        self.page_size = memory.shape[1]
        self.loaded: Queue = loaded
        super().__init__(daemon=True)

    def run(self):
        import hashlib
        with open(self.fname, 'rb') as handle:
            fileno = handle.fileno()

            while True:
                query = self.queries.get()
                # No more work
                if query is None:
                    break

                page_number, slot = query
                offset = np.uint64(page_number * self.page_size)
                length = read(fileno, self.memory[slot], offset)
                # print("L", page_number, slot, hashlib.md5(self.memory[slot]).hexdigest(), self.memory[slot].ctypes.data, length)
                self.loaded.put(page_number)




from ctypes import pointer
from tempfile import NamedTemporaryFile
from collections import defaultdict
from assertpy.assertpy import assert_that
from multiprocessing import cpu_count

import torch as ch
from assertpy import assert_that
import numpy as np
from torch.utils.data import Dataset
from ffcv import DatasetWriter
from ffcv.fields import IntField, NDArrayField, TorchTensorField
from ffcv import Loader

class DummyActivationsDataset(Dataset):

    def __init__(self, n_samples, shape, is_ch=False):
        self.n_samples = n_samples
        self.shape = shape
        self.is_ch = is_ch

    def __len__(self):
        return self.n_samples

    def __getitem__(self, index):
        if index >= self.n_samples:
            raise IndexError()
        np.random.seed(index)
        to_return = np.random.randn(*self.shape).astype('<f4')
        if self.is_ch:
            to_return = ch.from_numpy(to_return)

        return index, to_return

class TripleDummyActivationsDataset(Dataset):

    def __init__(self, n_samples, shape):
        self.n_samples = n_samples
        self.shape = shape

    def __len__(self):
        return self.n_samples

    def __getitem__(self, index):
        if index >= self.n_samples:
            raise IndexError()
        np.random.seed(index)
        d1 = np.random.randn(*self.shape).astype('<f4')
        d2 = np.random.randn(*self.shape).astype('<f4')
        d3 = np.random.randn(*self.shape).astype('<f4')
        return index, d1, d2, d3

def run_test(n_samples, shape, is_ch=False):
    with NamedTemporaryFile() as handle:
        name = handle.name
        dataset = DummyActivationsDataset(n_samples, shape, is_ch)

        if is_ch:
            field = TorchTensorField(ch.float32, shape)
        else:
            field = NDArrayField(np.dtype('<f4'), shape)

        writer = DatasetWriter(name, {
            'index': IntField(),
            'activations': field
        }, num_workers=3)

        writer.from_indexed_dataset(dataset)

        loader = Loader(name, batch_size=3, num_workers=min(5, cpu_count()))
        for ixes, activations in loader:
            for ix, activation in zip(ixes, activations):
                d = dataset[ix][1]
                if is_ch:
                    d = d.numpy()
                assert_that(np.all(d == activation.numpy())).is_true()


def test_simple_activations():
    run_test(4096, (2048,))

def test_simple_activations_ch():
    run_test(4096, (2048,), True)

def test_multi_fields():
    n_samples = 4096
    shape = (2048,)

    with NamedTemporaryFile() as handle:
        name = handle.name
        dataset = TripleDummyActivationsDataset(n_samples, shape)
        writer = DatasetWriter(name, {
            'index': IntField(),
            'activations': NDArrayField(np.dtype('<f4'), shape),
            'activations2': NDArrayField(np.dtype('<f4'), shape),
            'activations3': NDArrayField(np.dtype('<f4'), shape)
        }, num_workers=1)


        writer.from_indexed_dataset(dataset)

        loader = Loader(name, batch_size=3, num_workers=min(5, cpu_count()))
        page_size_l2 = int(np.log2(loader.reader.page_size))
        sample_ids = loader.reader.alloc_table['sample_id']
        pointers = loader.reader.alloc_table['ptr']
        pages = pointers >> page_size_l2
        sample_to_pages = defaultdict(set)

        for sample_id, page in zip(sample_ids, pages):
            sample_to_pages[sample_id].add(page)
            assert_that(sample_to_pages[sample_id]).is_length(1)

        for ixes, activations, d2, d3 in loader:
            for ix, activation in zip(ixes, activations):
                assert_that(np.all(dataset[ix][1] == activation.numpy())).is_true()


import ctypes
from numba import njit
import numpy as np
from ...libffcv import ctypes_resize

@njit(inline='always')
def resize_crop(source, start_row, end_row, start_col, end_col, destination):
    ctypes_resize(0,
                  source.ctypes.data,
                  source.shape[0], source.shape[1],
                  start_row, end_row, start_col, end_col,
                  destination.ctypes.data,
                  destination.shape[0], destination.shape[1])


@njit(parallel=False, fastmath=True, inline='always')
def get_random_crop(height, width, scale, ratio):
    area = height * width
    log_ratio = np.log(ratio)
    for _ in range(10):
        target_area = area * np.random.uniform(scale[0], scale[1])
        aspect_ratio = np.exp(np.random.uniform(log_ratio[0], log_ratio[1]))
        w = int(round(np.sqrt(target_area * aspect_ratio)))
        h = int(round(np.sqrt(target_area / aspect_ratio)))
        if 0 < w <= width and 0 < h <= height:
            i = int(np.random.uniform(0, height - h + 1))
            j = int(np.random.uniform(0, width - w + 1))
            return i, j, h, w
    in_ratio = float(width) / float(height)
    if in_ratio < min(ratio):
        w = width
        h = int(round(w / min(ratio)))
    elif in_ratio > max(ratio):
        h = height
        w = int(round(h * max(ratio)))
    else:
        w = width
        h = height
    i = (height - h) // 2
    j = (width - w) // 2
    return i, j, h, w


@njit(parallel=False, fastmath=True, inline='always')
def get_center_crop(height, width, ratio):
    s = min(height, width)
    c = int(ratio * s)
    delta_h = (height - c) // 2
    delta_w = (width - c) // 2

    return delta_h, delta_w, c, c



import string
from ctypes import pointer
from tempfile import NamedTemporaryFile
from collections import defaultdict
from multiprocessing import cpu_count

from assertpy import assert_that
import numpy as np
from torch.utils.data import Dataset
from ffcv import DatasetWriter
from ffcv.fields import IntField, JSONField
from ffcv.fields.bytes import BytesDecoder
from ffcv.fields.basics import IntDecoder
from ffcv import Loader

options = list(string.ascii_uppercase + string.digits)

def generate_random_string(low, high):
    length = np.random.randint(low, high)
    content = ''.join(np.random.choice(options, size=length))
    return content

class DummyDictDataset(Dataset):

    def __init__(self, n_samples):
        self.n_samples = n_samples

    def __len__(self):
        return self.n_samples

    def __getitem__(self, index):
        if index >= self.n_samples:
            raise IndexError()
        np.random.seed(index)
        length = np.random.randint(5, 250)
        content = np.random.randint(0, 256, size=(length,))
        json_content = {}
        for i in range(3):
            json_content[generate_random_string(5, 10)] = generate_random_string(50, 250)
        return index, json_content

def run_test(n_samples):
    with NamedTemporaryFile() as handle:
        name = handle.name
        dataset = DummyDictDataset(n_samples)
        writer = DatasetWriter(name, {
            'index': IntField(),
            'activations': JSONField()
        }, num_workers=min(3, cpu_count()))

        writer.from_indexed_dataset(dataset)

        loader = Loader(name, batch_size=3, num_workers=min(5, cpu_count()),
                        pipelines={
                            'activations': [BytesDecoder()],
                            'index': [IntDecoder()]
                        }
        )
        ix = 0
        for _, json_encoded in loader:
            json_docs = JSONField.unpack(json_encoded)
            for doc in json_docs:
                ref_doc = dataset[ix][1]
                assert_that(sorted(doc.items())).is_equal_to(sorted(ref_doc.items()))
                ix += 1


def test_simple_dict():
    run_test(32)

if __name__ == '__main__':
    test_simple_dict()


from typing import TYPE_CHECKING

import numpy as np
import numba as nb

from .base import MemoryManager, MemoryContext
from ..pipeline.compiler import Compiler

if TYPE_CHECKING:
    from ..reader import Reader


class OSCacheContext(MemoryContext):
    def __init__(self, manager:MemoryManager):
        self.manager = manager
        self.mmap = None

    @property
    def state(self):
        return (self.mmap, self.manager.ptrs, self.manager.sizes)

    def __enter__(self):
        res = super().__enter__()
        if self.mmap is None:
            self.mmap = np.memmap(self.manager.reader.file_name,
                                  'uint8', mode='r')
        return res

    def __exit__(self, __exc_type, __exc_value, __traceback):
        # Numpy doesn't have an API to close memory maps yet
        # The only thing one can do is flush it be since we are not
        # Writing to it it's pointless
        # Moreover we want to avoid opening the memmap over and over
        # anyway.
        return super().__exit__(__exc_type, __exc_value, __traceback)


class OSCacheManager(MemoryManager):

    def __init__(self, reader: 'Reader'):
        super().__init__(reader)
        self.context = OSCacheContext(self)

    def schedule_epoch(self, schedule):
        return self.context

    @property
    def state_type(self):
        t1 = nb.uint8[::1]
        t1.multable = False
        t2 = nb.uint64[::1]
        t1.mutable = False
        return nb.types.Tuple([t1, t2, t2])

    def compile_reader(self):
        def read(address, mem_state):
            size = mem_state[2][np.searchsorted(mem_state[1], address)]
            return mem_state[0][address:address + size]

        return Compiler.compile(read, nb.uint8[::1](nb.uint64, self.state_type))




from typing import Any, Optional, Sequence, Mapping

import torch as ch
import numpy as np

from .state import State
from .operation import Operation
from .allocation_query import Allocation, AllocationQuery

BAD_COLLATION_MESSAGE: str = "Each pipeline needs one and one only Collate operation"


class Pipeline:

    def __init__(self, operations: Sequence[Operation]):

        # This is the starting state of the pipeline
        self.original_state = State(jit_mode=True,
                                    device=ch.device('cpu'),
                                    dtype=np.dtype('u1'),
                                    shape=None)

        self.operations = operations

        self.operation_blocks, _ = self.parse_pipeline()
        self.compiled_ops = self.compile_ops()

        # Compile the pipeline
        self.compiled_code = None

    def parse_pipeline(self, batch_size=16):
        memory_allocations: Mapping[int, Optional[Allocation]] = {}
        operation_blocs = []

        current_state: State = self.original_state
        current_block = []

        # We read the content of the pipeline, validate and collect
        # Memory allocations
        for op_id, operation in enumerate(self.operations):
            previous_state = current_state
            current_state, memory_allocation = operation.declare_state_and_memory(
                current_state)

            if current_state.jit_mode != previous_state.jit_mode:
                if current_block:
                    operation_blocs.append((previous_state.jit_mode, current_block))
                current_block = [op_id]
            else:
                current_block.append(op_id)

            memory_allocations[op_id] = memory_allocation

        if current_block:
            operation_blocs.append((current_state.jit_mode, current_block))

        return operation_blocs, memory_allocations

    def compile_ops(self):
        compiled_ops = {}
        for op_id, operation in enumerate(self.operations):
            compiled_ops[op_id] = operation.generate_code()
        return compiled_ops

    def allocate_query(self, memory_allocation: AllocationQuery, batch_size: int, batches_ahead: int):
        # We compute the total amount of memory needed for this
        # operation
        final_shape = [batches_ahead,
                       batch_size, *memory_allocation.shape]
        if isinstance(memory_allocation.dtype, ch.dtype):
            result = []
            for _ in range(final_shape[0]):
                partial = ch.empty(*final_shape[1:],
                                  dtype=memory_allocation.dtype,
                                  device=memory_allocation.device)
                try:
                    partial = partial.pin_memory()
                except:
                    pass
                result.append(partial)
        else:
            ch_dtype = ch.from_numpy(np.empty(0, dtype=memory_allocation.dtype)).dtype
            result = ch.empty(*final_shape,
                              dtype=ch_dtype)
            try:
                result = result.pin_memory()
            except:
                pass
            result = result.numpy()
        return result


    def allocate_memory(self, batch_size: int, batches_ahead: int):
        _, memory_allocations = self.parse_pipeline()
        # Contains the actual allocated memory
        memory_buffers: Mapping[int, Any] = {}

        # For each allocation made by the operations in the pipeline
        for op_id, memory_allocation in memory_allocations.items():
            # If the operation didn't make a query we stop here
            allocated_buffer = None
            if isinstance(memory_allocation, AllocationQuery):
                allocated_buffer = self.allocate_query(memory_allocation,
                                                            batch_size,
                                                            batches_ahead)
            elif isinstance(memory_allocation, Sequence):
                allocated_buffer = tuple(
                    self.allocate_query(q, batch_size, batches_ahead) for q in memory_allocation
                )

            memory_buffers[op_id] = allocated_buffer

        return memory_buffers



"""
Image normalization
"""
from collections.abc import Sequence
from typing import Tuple

import numpy as np
import torch as ch
from numpy import dtype
from numpy.random import rand
from dataclasses import replace
from typing import Callable, Optional, Tuple
from ..pipeline.allocation_query import AllocationQuery
from ..pipeline.operation import Operation
from ..pipeline.state import State
from ..pipeline.compiler import Compiler

def ch_dtype_from_numpy(dtype):
    return ch.from_numpy(np.zeros((), dtype=dtype)).dtype

class NormalizeImage(Operation):
    """Fast implementation of normalization and type conversion for uint8 images
    to any floating point dtype.

    Works on both GPU and CPU tensors.

    Parameters
    ----------
    mean: np.ndarray
        The mean vector.
    std: np.ndarray
        The standard deviation vector.
    type: np.dtype
        The desired output type for the result as a numpy type.
        If the transform is applied on a GPU tensor it will be converted
        as the equivalent torch dtype.
    """

    def __init__(self, mean: np.ndarray, std: np.ndarray,
                 type: np.dtype):
        super().__init__()
        table = (np.arange(256)[:, None] - mean[None, :]) / std[None, :]
        self.original_dtype = type
        table = table.astype(type)
        if type == np.float16:
            type = np.int16
        self.dtype = type
        table = table.view(type)
        self.lookup_table = table
        self.previous_shape = None
        self.mode = 'cpu'

    def generate_code(self) -> Callable:
        if self.mode == 'cpu':
            return self.generate_code_cpu()
        return self.generate_code_gpu()

    def generate_code_gpu(self) -> Callable:

        # We only import cupy if it's truly needed
        import cupy as cp
        import pytorch_pfn_extras as ppe

        tn = np.zeros((), dtype=self.dtype).dtype.name
        kernel = cp.ElementwiseKernel(f'uint8 input, raw {tn} table', f'{tn} output', 'output = table[input * 3 + i % 3];')
        final_type = ch_dtype_from_numpy(self.original_dtype)
        s = self
        def normalize_convert(images, result):
            B, C, H, W = images.shape
            table = self.lookup_table.view(-1)
            assert images.is_contiguous(memory_format=ch.channels_last), 'Images need to be in channel last'
            result = result[:B]
            result_c = result.view(-1)
            images = images.permute(0, 2, 3, 1).view(-1)

            current_stream = ch.cuda.current_stream()
            with ppe.cuda.stream(current_stream):
                kernel(images, table, result_c)

            # Mark the result as channel last
            final_result = result.reshape(B, H, W, C).permute(0, 3, 1, 2)

            assert final_result.is_contiguous(memory_format=ch.channels_last), 'Images need to be in channel last'

            return final_result.view(final_type)

        return normalize_convert

    def generate_code_cpu(self) -> Callable:

        table = self.lookup_table.view(dtype=self.dtype)
        my_range = Compiler.get_iterator()

        def normalize_convert(images, result, indices):
            result_flat = result.reshape(result.shape[0], -1, 3)
            num_pixels = result_flat.shape[1]
            for i in my_range(len(indices)):
                image = images[i].reshape(num_pixels, 3)
                for px in range(num_pixels):
                    # Just in case llvm forgets to unroll this one
                    result_flat[i, px, 0] = table[image[px, 0], 0]
                    result_flat[i, px, 1] = table[image[px, 1], 1]
                    result_flat[i, px, 2] = table[image[px, 2], 2]

            return result

        normalize_convert.is_parallel = True
        normalize_convert.with_indices = True
        return normalize_convert

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:

        if previous_state.device == ch.device('cpu'):
            new_state = replace(previous_state, jit_mode=True, dtype=self.dtype)
            return new_state, AllocationQuery(
                shape=previous_state.shape,
                dtype=self.dtype,
                device=previous_state.device
            )

        else:
            self.mode = 'gpu'
            new_state = replace(previous_state, dtype=self.dtype)

            gpu_type = ch_dtype_from_numpy(self.dtype)


            # Copy the lookup table into the proper device
            try:
                self.lookup_table = ch.from_numpy(self.lookup_table)
            except TypeError:
                pass  # This is alredy a tensor
            self.lookup_table = self.lookup_table.to(previous_state.device)

            return new_state, AllocationQuery(
                shape=previous_state.shape,
                device=previous_state.device,
                dtype=gpu_type
            )


import numpy as np
from assertpy import assert_that
from torch.utils.data import Dataset
import logging
import os
from assertpy import assert_that
from tempfile import NamedTemporaryFile

from ffcv.writer import DatasetWriter
from ffcv.reader import Reader
from ffcv.fields import BytesField, IntField

class DummyDataset(Dataset):

    def __init__(self, l, size):
        self.l = l
        self.size = size

    def __len__(self):
        return self.l

    def __getitem__(self, index):
        if index > self.l:
            raise IndexError
        np.random.seed(index)
        return index, np.random.randint(0, 255, size=self.size, dtype='u1')


def create_and_validate(length, size):

    dataset = DummyDataset(length, size)

    with NamedTemporaryFile() as handle:
        name = handle.name
        writer = DatasetWriter(name, {
            'index': IntField(),
            'value': BytesField()
        }, num_workers=2)

        writer.from_indexed_dataset(dataset, chunksize=5)

        reader = Reader(name)

        assert_that(reader.handlers).is_length(2)
        assert_that(reader.handlers['index']).is_instance_of(IntField)
        assert_that(reader.handlers['value']).is_instance_of(BytesField)
        assert_that(reader.alloc_table).is_length(length)
        assert_that(reader.metadata).is_length(length)
        assert_that((reader.metadata['f0'] == np.arange(length).astype('int')).all()).is_true()

        assert_that(np.all(reader.alloc_table['size'] == size)).is_true()

def test_simple():
    create_and_validate(600, 76)

def test_large():
    create_and_validate(600, 1024)

def test_many():
    create_and_validate(60000, 81)





from os import path

import numpy as np
import cv2
from numpy.core.numeric import full

from ..decorator import benchmark
from ..benchmark import Benchmark

from ...pipeline.compiler import Compiler

from ...libffcv import imdecode

@benchmark({
    'n': [500],
    'source_image': ['../../../test_data/pig.png'],
    'image_width': [500, 256, 1024],
    'quality': [50, 90],
    'compile': [True]
})
class JPEGDecodeBenchmark(Benchmark):

    def __init__(self, n, source_image, image_width, quality, compile):
        self.n = n
        self.compile = compile
        self.source_image = source_image
        self.image_width = image_width
        self.quality = quality

    def __enter__(self):
        full_path = path.join(path.dirname(__file__), self.source_image)
        loaded_image = cv2.imread(full_path, cv2.IMREAD_COLOR)   
        previous_width = loaded_image.shape[1]
        new_width = self.image_width
        factor = new_width / previous_width
        new_height = int(loaded_image.shape[0] * factor)
        resized_image = cv2.resize(loaded_image, (new_width, new_height),
                                   interpolation=cv2.INTER_AREA)
        _, self.encoded_image = cv2.imencode('.jpg', resized_image,
                                          [int(cv2.IMWRITE_JPEG_QUALITY),
                                              self.quality])
                                          
        self.destination = np.zeros((new_height, new_width, 3), dtype='uint8')
        
        Compiler.set_enabled(self.compile)
        
        n = self.n
        decode = Compiler.compile(imdecode)
        def code(source, dest):
            for _ in range(n):
                decode(source, dest,
                           new_height,
                           new_width,
                           new_height,
                           new_width,
                           0, 0, 1, 1, False)
                
        self.code = Compiler.compile(code)
        
    def run(self):
        self.code(self.encoded_image, self.destination)
        
    def __exit__(self, *args):
        pass



from abc import ABCMeta, abstractmethod
from contextlib import AbstractContextManager

class Benchmark(AbstractContextManager, metaclass=ABCMeta):
    
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def run(self):
        raise NotImplemented()


from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
from typing import Callable, Optional, Tuple

import numpy as np

from .state import State
from .allocation_query import AllocationQuery

if TYPE_CHECKING:
    from ..fields.base import Field


class Operation(ABC):

    def __init__(self):
        self.metadata: np.ndarray = None
        self.memory_read: Callable[[np.uint64], np.ndarray] = None
        pass

    def accept_field(self, field: 'Field'):
        self.field: 'Field' = field

    def accept_globals(self, metadata, memory_read):
        self.metadata = metadata
        self.memory_read = memory_read

    # Return the code to run this operation
    @abstractmethod
    def generate_code(self) -> Callable:
        raise NotImplementedError

    def declare_shared_memory(self, previous_state: State) -> Optional[AllocationQuery]:
        return None

    def generate_code_for_shared_state(self) -> Optional[Callable]:
        return None

    @abstractmethod
    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        raise NotImplementedError



import numpy as np

from .utils import decode_null_terminated_string
from .types import (ALLOC_TABLE_TYPE, HeaderType, CURRENT_VERSION,
                    FieldDescType, get_handlers, get_metadata_type)

class Reader:

    def __init__(self, fname, custom_handlers={}):
        self._fname = fname
        self._custom_handlers = custom_handlers
        self.read_header()
        self.read_field_descriptors()
        self.read_metadata()
        self.read_allocation_table()

    @property
    def file_name(self):
        return self._fname

    def read_header(self):
        header = np.fromfile(self._fname, dtype=HeaderType, count=1)[0]
        header.setflags(write=False)
        version = header['version']

        if version != CURRENT_VERSION:
            msg = f"file format mismatch: code={CURRENT_VERSION},file={version}"
            raise AssertionError(msg)

        self.num_samples = header['num_samples']
        self.page_size = header['page_size']
        self.num_fields = header['num_fields']
        self.header = header

    def read_field_descriptors(self):
        offset = HeaderType.itemsize
        field_descriptors = np.fromfile(self._fname, dtype=FieldDescType,
                                        count=self.num_fields, offset=offset)
        field_descriptors.setflags(write=False)
        handlers = get_handlers(field_descriptors)

        self.field_descriptors = field_descriptors
        self.field_names = list(map(decode_null_terminated_string,
                                    self.field_descriptors['name']))
        self.handlers = dict(zip(self.field_names, handlers))

        for field_name, field_desc in zip(self.field_names, self.field_descriptors):
            if field_name in self._custom_handlers:
                CustomHandler = self._custom_handlers[field_name]
                self.handlers[field_name] = CustomHandler.from_binary(field_desc['arguments'])
        
        for field_name, handler in self.handlers.items():
            if handler is None:
                raise ValueError(f"Must specify a custom_field entry " \
                                 f"for custom field {field_name}")

        self.metadata_type = get_metadata_type(list(self.handlers.values()))

    def read_metadata(self):
        offset = HeaderType.itemsize + self.field_descriptors.nbytes
        self.metadata = np.fromfile(self._fname, dtype=self.metadata_type,
                                   count=self.num_samples, offset=offset)
        self.metadata.setflags(write=False)

    def read_allocation_table(self):
        offset = self.header['alloc_table_ptr']
        alloc_table = np.fromfile(self._fname, dtype=ALLOC_TABLE_TYPE,
                                  offset=offset)
        alloc_table.setflags(write=False)
        self.alloc_table = alloc_table





from dataclasses import replace
import torch as ch
from ffcv.pipeline.allocation_query import AllocationQuery
from ffcv.pipeline.compiler import Compiler
import numpy as np
from typing import Callable
from assertpy import assert_that
from torch.utils.data import Dataset
import logging
import os
from assertpy import assert_that
from tempfile import NamedTemporaryFile
from ffcv.pipeline.operation import Operation
from ffcv.transforms.ops import ToTensor
from multiprocessing import cpu_count

from ffcv.writer import DatasetWriter
from ffcv.reader import Reader
from ffcv.loader import Loader
from ffcv.fields import IntField, FloatField, BytesField
from ffcv.fields.basics import FloatDecoder
from ffcv.pipeline.state import State

from test_writer import DummyDataset

numba_logger = logging.getLogger('numba')
numba_logger.setLevel(logging.WARNING)


class Doubler(Operation):

    def generate_code(self) -> Callable:
        def code(x, dst):
            dst[:] = x * 2
            return dst
        return code

    def declare_state_and_memory(self, previous_state: State):
        return (previous_state, AllocationQuery(previous_state.shape, previous_state.dtype, previous_state.device))

def test_basic_simple():
    length = 600
    batch_size = 8
    with NamedTemporaryFile() as handle:
        file_name = handle.name
        dataset = DummyDataset(length)
        writer = DatasetWriter(file_name, {
            'index': IntField(),
            'value': FloatField()
        })

        writer.from_indexed_dataset(dataset)

        Compiler.set_enabled(True)

        loader = Loader(file_name, batch_size, num_workers=min(5, cpu_count()), seed=17,
                        pipelines={
                            'value': [FloatDecoder(), Doubler(), ToTensor()]
                        })

        it = iter(loader)
        indices, values = next(it)
        assert_that(np.allclose(indices.squeeze().numpy(),
                                np.arange(batch_size))).is_true()
        assert_that(np.allclose(2 * np.sin(np.arange(batch_size)),
                                values.squeeze().numpy())).is_true()

def test_multiple_iterators_success():
    length = 60
    batch_size = 8
    with NamedTemporaryFile() as handle:
        file_name = handle.name
        dataset = DummyDataset(length)
        writer = DatasetWriter(file_name, {
            'index': IntField(),
            'value': FloatField()
        })

        writer.from_indexed_dataset(dataset)

        Compiler.set_enabled(True)

        loader = Loader(file_name, batch_size, num_workers=min(5, cpu_count()), seed=17,
                        pipelines={
                            'value': [FloatDecoder(), Doubler(), ToTensor()]
                        })

        it = iter(loader)
        it = iter(loader)

def test_multiple_epoch_doesnt_recompile():
    length = 60
    batch_size = 8
    with NamedTemporaryFile() as handle:
        file_name = handle.name
        dataset = DummyDataset(length)
        writer = DatasetWriter(file_name, {
            'index': IntField(),
            'value': FloatField()
        })

        writer.from_indexed_dataset(dataset)

        Compiler.set_enabled(True)

        loader = Loader(file_name, batch_size, num_workers=min(5, cpu_count()), seed=17,
                        pipelines={
                            'value': [FloatDecoder(), Doubler(), ToTensor()]
                        })

        it = iter(loader)
        code = loader.code
        it = iter(loader)
        new_code = loader.code
        assert_that(code).is_equal_to(new_code)

def test_multiple_epoch_does_recompile():
    length = 60
    batch_size = 8
    with NamedTemporaryFile() as handle:
        file_name = handle.name
        dataset = DummyDataset(length)
        writer = DatasetWriter(file_name, {
            'index': IntField(),
            'value': FloatField()
        })

        writer.from_indexed_dataset(dataset)

        Compiler.set_enabled(True)

        loader = Loader(file_name, batch_size, num_workers=min(5, cpu_count()), seed=17,
                recompile=True,
                        pipelines={
                            'value': [FloatDecoder(), Doubler(), ToTensor()]
                        })

        it = iter(loader)
        code = loader.code
        it = iter(loader)
        new_code = loader.code
        assert_that(code).is_not_equal_to(new_code)


"""
Fast training script for CIFAR-10 using FFCV.
For tutorial, see https://docs.ffcv.io/ffcv_examples/cifar10.html.

First, from the same directory, run:

    `python write_datasets.py --data.train_dataset [TRAIN_PATH] \
                              --data.val_dataset [VAL_PATH]`

to generate the FFCV-formatted versions of CIFAR.

Then, simply run this to train models with default hyperparameters:

    `python train_cifar.py --config-file default_config.yaml`

You can override arguments as follows:

    `python train_cifar.py --config-file default_config.yaml \
                           --training.lr 0.2 --training.num_workers 4 ... [etc]`

or by using a different config file.
"""
from argparse import ArgumentParser
from typing import List
import time
import numpy as np
from tqdm import tqdm

import torch as ch
from torch.cuda.amp import GradScaler, autocast
from torch.nn import CrossEntropyLoss, Conv2d, BatchNorm2d
from torch.optim import SGD, lr_scheduler
import torchvision

from fastargs import get_current_config, Param, Section
from fastargs.decorators import param
from fastargs.validation import And, OneOf

from ffcv.fields import IntField, RGBImageField
from ffcv.fields.decoders import IntDecoder, SimpleRGBImageDecoder
from ffcv.loader import Loader, OrderOption
from ffcv.pipeline.operation import Operation
from ffcv.transforms import RandomHorizontalFlip, Cutout, \
    RandomTranslate, Convert, ToDevice, ToTensor, ToTorchImage
from ffcv.transforms.common import Squeeze
from ffcv.writer import DatasetWriter

Section('training', 'Hyperparameters').params(
    lr=Param(float, 'The learning rate to use', required=True),
    epochs=Param(int, 'Number of epochs to run for', required=True),
    lr_peak_epoch=Param(int, 'Peak epoch for cyclic lr', required=True),
    batch_size=Param(int, 'Batch size', default=512),
    momentum=Param(float, 'Momentum for SGD', default=0.9),
    weight_decay=Param(float, 'l2 weight decay', default=5e-4),
    label_smoothing=Param(float, 'Value of label smoothing', default=0.1),
    num_workers=Param(int, 'The number of workers', default=8),
    lr_tta=Param(bool, 'Test time augmentation by averaging with horizontally flipped version', default=True)
)

Section('data', 'data related stuff').params(
    train_dataset=Param(str, '.dat file to use for training', required=True),
    val_dataset=Param(str, '.dat file to use for validation', required=True),
)

@param('data.train_dataset')
@param('data.val_dataset')
@param('training.batch_size')
@param('training.num_workers')
def make_dataloaders(train_dataset=None, val_dataset=None, batch_size=None, num_workers=None):
    paths = {
        'train': train_dataset,
        'test': val_dataset

    }

    start_time = time.time()
    CIFAR_MEAN = [125.307, 122.961, 113.8575]
    CIFAR_STD = [51.5865, 50.847, 51.255]
    loaders = {}

    for name in ['train', 'test']:
        label_pipeline: List[Operation] = [IntDecoder(), ToTensor(), ToDevice(ch.device('cuda:0')), Squeeze()]
        image_pipeline: List[Operation] = [SimpleRGBImageDecoder()]
        if name == 'train':
            image_pipeline.extend([
                RandomHorizontalFlip(),
                RandomTranslate(padding=2, fill=tuple(map(int, CIFAR_MEAN))),
                Cutout(4, tuple(map(int, CIFAR_MEAN))),
            ])
        image_pipeline.extend([
            ToTensor(),
            ToDevice(ch.device('cuda:0'), non_blocking=True),
            ToTorchImage(),
            Convert(ch.float16),
            torchvision.transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
        ])
        
        ordering = OrderOption.RANDOM if name == 'train' else OrderOption.SEQUENTIAL

        loaders[name] = Loader(paths[name], batch_size=batch_size, num_workers=num_workers,
                               order=ordering, drop_last=(name == 'train'),
                               pipelines={'image': image_pipeline, 'label': label_pipeline})

    return loaders, start_time

# Model (from KakaoBrain: https://github.com/wbaek/torchskeleton)
class Mul(ch.nn.Module):
    def __init__(self, weight):
       super(Mul, self).__init__()
       self.weight = weight
    def forward(self, x): return x * self.weight

class Flatten(ch.nn.Module):
    def forward(self, x): return x.view(x.size(0), -1)

class Residual(ch.nn.Module):
    def __init__(self, module):
        super(Residual, self).__init__()
        self.module = module
    def forward(self, x): return x + self.module(x)

def conv_bn(channels_in, channels_out, kernel_size=3, stride=1, padding=1, groups=1):
    return ch.nn.Sequential(
            ch.nn.Conv2d(channels_in, channels_out, kernel_size=kernel_size,
                         stride=stride, padding=padding, groups=groups, bias=False),
            ch.nn.BatchNorm2d(channels_out),
            ch.nn.ReLU(inplace=True)
    )

def construct_model():
    num_class = 10
    model = ch.nn.Sequential(
        conv_bn(3, 64, kernel_size=3, stride=1, padding=1),
        conv_bn(64, 128, kernel_size=5, stride=2, padding=2),
        Residual(ch.nn.Sequential(conv_bn(128, 128), conv_bn(128, 128))),
        conv_bn(128, 256, kernel_size=3, stride=1, padding=1),
        ch.nn.MaxPool2d(2),
        Residual(ch.nn.Sequential(conv_bn(256, 256), conv_bn(256, 256))),
        conv_bn(256, 128, kernel_size=3, stride=1, padding=0),
        ch.nn.AdaptiveMaxPool2d((1, 1)),
        Flatten(),
        ch.nn.Linear(128, num_class, bias=False),
        Mul(0.2)
    )
    model = model.to(memory_format=ch.channels_last).cuda()
    return model

@param('training.lr')
@param('training.epochs')
@param('training.momentum')
@param('training.weight_decay')
@param('training.label_smoothing')
@param('training.lr_peak_epoch')
def train(model, loaders, lr=None, epochs=None, label_smoothing=None,
          momentum=None, weight_decay=None, lr_peak_epoch=None):
    opt = SGD(model.parameters(), lr=lr, momentum=momentum, weight_decay=weight_decay)
    iters_per_epoch = len(loaders['train'])
    # Cyclic LR with single triangle
    lr_schedule = np.interp(np.arange((epochs+1) * iters_per_epoch),
                            [0, lr_peak_epoch * iters_per_epoch, epochs * iters_per_epoch],
                            [0, 1, 0])
    scheduler = lr_scheduler.LambdaLR(opt, lr_schedule.__getitem__)
    scaler = GradScaler()
    loss_fn = CrossEntropyLoss(label_smoothing=label_smoothing)

    for _ in range(epochs):
        for ims, labs in tqdm(loaders['train']):
            opt.zero_grad(set_to_none=True)
            with autocast():
                out = model(ims)
                loss = loss_fn(out, labs)

            scaler.scale(loss).backward()
            scaler.step(opt)
            scaler.update()
            scheduler.step()

@param('training.lr_tta')
def evaluate(model, loaders, lr_tta=False):
    model.eval()
    with ch.no_grad():
        for name in ['train', 'test']:
            total_correct, total_num = 0., 0.
            for ims, labs in tqdm(loaders[name]):
                with autocast():
                    out = model(ims)
                    if lr_tta:
                        out += model(ims.flip(-1))
                    total_correct += out.argmax(1).eq(labs).sum().cpu().item()
                    total_num += ims.shape[0]
            print(f'{name} accuracy: {total_correct / total_num * 100:.1f}%')

if __name__ == "__main__":
    config = get_current_config()
    parser = ArgumentParser(description='Fast CIFAR-10 training')
    config.augment_argparse(parser)
    # Also loads from args.config_path if provided
    config.collect_argparse_args(parser)
    config.validate(mode='stderr')
    config.summary()

    loaders, start_time = make_dataloaders()
    model = construct_model()
    train(model, loaders)
    print(f'Total time: {time.time() - start_time:.5f}')
    evaluate(model, loaders)



"""
Wrapper for a torch.nn.Module
"""
import torch as ch
from numpy.random import permutation, rand
from typing import Callable, Optional, Tuple
from ..pipeline.allocation_query import AllocationQuery
from ..pipeline.operation import Operation
from ..pipeline.state import State

class ModuleWrapper(Operation):
    """Transform using the given torch.nn.Module

    Parameters
    ----------
    module: torch.nn.Module
        The module for transformation
    """
    def __init__(self, module: ch.nn.Module):
        super().__init__()
        self.module = module

    def generate_code(self) -> Callable:
        def apply_module(inp, _):
            res = self.module(inp)
            return res

        return apply_module

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        return previous_state, None



from os import path
from glob import glob
import tempfile

import numpy as np
from tempfile import TemporaryDirectory, NamedTemporaryFile
import torch as ch
from torch.utils.data import Dataset
import webdataset as wds

from ffcv import DatasetWriter
from ffcv.reader import Reader
from ffcv.fields import IntField, FloatField
from test_writer import validate_simple_dataset

field_names = [
    'index',
    'value.pyd'
]

class DummyDataset(Dataset):

    def __init__(self, l):
        self.l = l

    def __len__(self):
        return self.l

    def __getitem__(self, index):
        if index >= self.l:
            raise IndexError()
        return (index, np.sin(index))

def write_webdataset(folder, dataset, field_names):
    pattern = path.join(folder, "dataset-%06d.tar")
    writer = wds.ShardWriter(pattern, maxcount=20)
    with writer as sink:
        for i, sample in enumerate(dataset):
            data = {
                '__key__': f'sample_{i}'
            }

            for field_name, value in zip(field_names, sample):
                data[field_name] = value
            sink.write(data)


def pipeline(dataset):
    return (dataset
        .decode()
        .to_tuple(*field_names)
    )

if __name__ == '__main__':
    N = 1007
    dataset = DummyDataset(N)
    with TemporaryDirectory() as temp_directory:
        with NamedTemporaryFile() as handle:
            fname = handle.name
            write_webdataset(temp_directory, dataset, field_names)
            files = glob(path.join(temp_directory, '*'))
            files = list(sorted(files))

            print(fname)
            writer = DatasetWriter(fname, {
                'index': IntField(),
                'value': FloatField()
            })

            writer.from_webdataset(files, pipeline)

            validate_simple_dataset(fname, N, shuffled=False)


import numpy as np
from numba import types
from numba.extending import intrinsic


def chunks(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def is_power_of_2(n):
    return (n & (n-1) == 0) and n != 0

def align_to_page(ptr, page_size):
    # If we are not aligned with the start of a page:
    if ptr % page_size != 0:
        ptr = ptr  + page_size - ptr % page_size
    return ptr

def decode_null_terminated_string(bytes: np.ndarray):
    return bytes.tobytes().decode('ascii').split('\x00')[0]

@intrinsic
def cast_int_to_byte_ptr(typingctx, src):
    # check for accepted types
    if isinstance(src, types.Integer):
        # create the expected type signature
        result_type = types.CPointer(types.uint8)
        sig = result_type(types.uintp)
        # defines the custom code generation
        def codegen(context, builder, signature, args):
            # llvm IRBuilder code here
            [src] = args
            rtype = signature.return_type
            llrtype = context.get_value_type(rtype)
            return builder.inttoptr(src, llrtype)
        return sig, codegen
        
from threading import Lock
s_print_lock = Lock()


def s_print(*a, **b):
    """Thread safe print function"""
    with s_print_lock:
        print(*a, **b)
    


import logging
import os
from tempfile import NamedTemporaryFile
from time import sleep, time

import numpy as np
from assertpy import assert_that
from ffcv.fields import BytesField, IntField, RGBImageField
from ffcv.memory_managers import OSCacheManager
from ffcv.pipeline.compiler import Compiler
from ffcv.reader import Reader
from ffcv.writer import DatasetWriter
from torch.utils.data import Dataset
from tqdm import tqdm

from ..benchmark import Benchmark
from ..decorator import benchmark



class DummyDataset(Dataset):

    def __init__(self, length, size):
        self.length = length
        self.size = size
        
    def __len__(self):
        return self.length

    def __getitem__(self, index):
        if index > self.length:
            raise IndexError

        dims = tuple([*self.size, 3])
        image_data = np.random.randint(low=0, high=255, size=dims, dtype='uint8')
        return index, image_data



@benchmark({
    'n': [3000],
    'length': [3000],
    'mode': [
        'raw',
        'jpg'
        ],
    'num_workers': [
        1,
        8,
        16
    ],
    'batch_size': [
        500
    ],
    'size': [
        (32, 32),  # CIFAR
        (300, 500),  # ImageNet
    ],
    'compile': [
        True,
        # False
    ],
    'random_reads': [
        True,
        # False
    ]
})
class ImageReadBench(Benchmark):
    
    def __init__(self, n, length, mode, size, random_reads, compile, num_workers, batch_size):
        self.n = n
        self.mode = mode
        self.length = length
        self.num_workers = num_workers
        self.batch_size = batch_size
        self.size = size
        self.compile = compile
        self.random_reads = random_reads
        self.dataset = DummyDataset(length, size)
        
    def __enter__(self):
        self.handle = NamedTemporaryFile()
        self.handle.__enter__()
        name = self.handle.name

        writer = DatasetWriter(self.length, name, {
            'index': IntField(),
            'value': RGBImageField(write_mode=self.mode)
        })

        with writer:
            writer.write_pytorch_dataset(self.dataset, num_workers=-1, chunksize=100)

        reader = Reader(name)
        manager = OSCacheManager(reader)

        Compiler.set_enabled(self.compile)
        Compiler.set_num_threads(self.num_workers)

        memreader = manager.compile_reader()
        Decoder = RGBImageField().get_decoder_class()
        decoder = Decoder()
        decoder.accept_globals(reader.metadata['f1'], memreader)

        context = manager.schedule_epoch(np.arange(self.n))
        context.__enter__()
        self.context = context

        decode = decoder.generate_code()
        decode = Compiler.compile(decode)
        
        self.buff = np.zeros((self.batch_size, *self.size, 3), dtype='uint8')
        
        if self.random_reads:
            self.indices = np.random.choice(self.n, size=self.n, replace=False)
        else:
            self.indices = np.arange(self.n)
            
        def code(indices, buff, state):
            result = 0
            for i in range(0, len(indices), self.batch_size):
                result += decode(indices[i:i + self.batch_size], buff, reader.metadata['f1'], state)[0, 5, 5]
            return result
                
        self.code = code

    def run(self):
        self.code(self.indices, self.buff, self.context.state)

    def __exit__(self, *args):
        self.handle.__exit__(*args)
        self.context.__exit__(*args)
        pass



from setuptools import find_packages
import subprocess
from difflib import get_close_matches
from glob import glob
import os
import platform

from distutils.core import setup, Extension

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


def find_pkg_dirs(package):
    close_matches = get_close_matches(package.lower(),
                                      os.environ["PATH"].lower().split(';'),
                                      cutoff=0)
    dll_dir = None
    for close_match in close_matches:
        if (os.path.exists(close_match)
                and glob(os.path.join(close_match, '*.dll'))):
            dll_dir = close_match
            break
    if dll_dir is None:
        raise Exception(
            f"Could not find required package: {package}. "
            "Add directory containing .dll files to system environment path."
        )
    dll_dir_split = dll_dir.replace('\\', '/').split('/')
    root = get_close_matches(package.lower(), dll_dir_split, cutoff=0)[0]
    root_dir = '/'.join(dll_dir_split[:dll_dir_split.index(root) + 1])
    return os.path.normpath(root_dir), os.path.normpath(dll_dir)


def pkgconfig_windows(package, kw):
    is_x64 = platform.machine().endswith('64')
    root_dir, dll_dir = find_pkg_dirs(package)
    include_dir = None
    library_dir = None
    parent = None
    while parent != root_dir:
        parent = os.path.dirname(dll_dir if parent is None else parent)
        if include_dir is None and os.path.exists(os.path.join(parent, 'include')):
            include_dir = os.path.join(parent, 'include')
        library_dirs = set()
        libraries = glob(os.path.join(parent, '**', 'lib', '**', '*.lib'),
                         recursive=True)
        for library in libraries:
            if ((is_x64 and 'x86' in library)
                    or (not is_x64 and 'x64' in library)):
                continue
            library_dirs.add(os.path.dirname(library))
        if library_dir is None and library_dirs:
            library_dir = sorted(library_dirs)[-1]
        if include_dir and library_dir:
            libraries = [os.path.splitext(library)[0]
                         for library in glob(os.path.join(library_dir, '*.lib'))]
            break
    if not include_dir or not library_dir:
        raise Exception(f"Could not find required package: {package}.")
    kw.setdefault('include_dirs', []).append(include_dir)
    kw.setdefault('library_dirs', []).append(library_dir)
    kw.setdefault('libraries', []).extend(libraries)
    return kw


def pkgconfig(package, kw):
    flag_map = {'-I': 'include_dirs', '-L': 'library_dirs', '-l': 'libraries'}
    output = subprocess.getoutput(
        'pkg-config --cflags --libs {}'.format(package))
    if 'not found' in output:
        raise RuntimeError(f"Could not find required package: {package}.")
    for token in output.strip().split():
        kw.setdefault(flag_map.get(token[:2]), []).append(token[2:])
    return kw


sources = ['./libffcv/libffcv.cpp']

extension_kwargs = {
    'sources': sources,
    'include_dirs': []
}
if platform.system() == 'Windows':
    extension_kwargs = pkgconfig_windows('opencv4', extension_kwargs)
    extension_kwargs = pkgconfig_windows('libturbojpeg', extension_kwargs)

    extension_kwargs = pkgconfig_windows('pthread', extension_kwargs)
else:
    try:
        extension_kwargs = pkgconfig('opencv4', extension_kwargs)
    except RuntimeError:
        extension_kwargs = pkgconfig('opencv', extension_kwargs)
    extension_kwargs = pkgconfig('libturbojpeg', extension_kwargs)

    extension_kwargs['libraries'].append('pthread')


libffcv = Extension('ffcv._libffcv',
                        **extension_kwargs)

setup(name='ffcv',
      version='1.0.1',
      description=' FFCV: Fast Forward Computer Vision ',
      author='MadryLab',
      author_email='ffcv@mit.edu',
      url='https://github.com/libffcv/ffcv',
      license_files = ('LICENSE.txt',),
      packages=find_packages(),
      long_description=long_description,
      long_description_content_type='text/markdown',
      ext_modules=[libffcv],
      install_requires=[
          'terminaltables',
          'pytorch_pfn_extras',
          'fastargs',
          'opencv-python',
          'assertpy',
          'tqdm',
          'psutil',
          'numba',
      ])



from .basics import FloatDecoder, IntDecoder
from .ndarray import NDArrayDecoder
from .rgb_image import RandomResizedCropRGBImageDecoder, CenterCropRGBImageDecoder, SimpleRGBImageDecoder
from .bytes import BytesDecoder

__all__ = ['FloatDecoder', 'IntDecoder', 'NDArrayDecoder', 'RandomResizedCropRGBImageDecoder', 
           'CenterCropRGBImageDecoder', 'SimpleRGBImageDecoder', 'BytesDecoder']


from typing import Callable, TYPE_CHECKING, Tuple, Type
from dataclasses import replace

import numpy as np

from .base import Field, ARG_TYPE
from ..pipeline.operation import Operation
from ..pipeline.state import State
from ..pipeline.compiler import Compiler
from ..pipeline.allocation_query import AllocationQuery
from ..libffcv import memcpy


class BytesDecoder(Operation):

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, AllocationQuery]:
        max_size = self.metadata['size'].max()

        my_shape = (max_size,)
        return (
            replace(previous_state, jit_mode=True, shape=my_shape,
                    dtype='<u1'),
            AllocationQuery(my_shape, dtype='<u1')
        )

    def generate_code(self) -> Callable:
        mem_read = self.memory_read
        my_memcpy = Compiler.compile(memcpy)
        my_range = Compiler.get_iterator()
        def decoder(batch_indices, destination, metadata, storage_state):
            for dest_ix in my_range(batch_indices.shape[0]):
                source_ix = batch_indices[dest_ix]
                data = mem_read(metadata[source_ix]['ptr'], storage_state)
                my_memcpy(data, destination[dest_ix])
            return destination

        return decoder

class BytesField(Field):
    """
    A subclass of :class:`~ffcv.fields.Field` supporting variable-length byte
    arrays.

    Intended for use with data such as text or raw data which may not have a
    fixed size. Data is written sequentially while saving pointers and read by
    pointer lookup.

    The writer expects to be passed a 1D uint8 numpy array of variable length for each sample.
    """
    def __init__(self):
        pass

    @property
    def metadata_type(self) -> np.dtype:
        return np.dtype([
            ('ptr', '<u8'),
            ('size', '<u8')
        ])

    @staticmethod
    def from_binary(binary: ARG_TYPE) -> Field:
        return BytesField()

    def to_binary(self) -> ARG_TYPE:
        return np.zeros(1, dtype=ARG_TYPE)[0]

    def encode(self, destination, field, malloc):
        ptr, buffer = malloc(field.size)
        buffer[:] = field
        destination['ptr'] = ptr
        destination['size'] = field.size

    def get_decoder_class(self) -> Type[Operation]:
        return BytesDecoder



from __future__ import annotations
from typing import Type
import numpy as np
from abc import ABC, abstractmethod

from ..pipeline.operation import Operation

ARG_TYPE = np.dtype([('', '<u1', 1024)])

class Field(ABC):
    """
    Abstract Base Class for implementing fields (e.g., images, integers).

    Each dataset entry is comprised of one or more fields (for example, standard
    object detection datasets may have one field for images, and one for
    bounding boxes). Fields are responsible for implementing encode and get_decoder_class
    functions that determine how they will be written and read from the dataset file,
    respectively.

    .. note ::
        It is possible to have multiple potential decoder for a given Field. ``RGBImageField`` one example. Users can specify which decoder to use in the ``piplines`` argument of the ``Loader`` class.

    See :ref:`here <TODO>` for information on how to implement a subclass of Field.
    """
    @property
    @abstractmethod
    def metadata_type(self) -> np.dtype:
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def from_binary(binary: ARG_TYPE) -> Field:
        raise NotImplementedError

    @abstractmethod
    def to_binary(self) -> ARG_TYPE:
        raise NotImplementedError

    @abstractmethod
    def encode(field, metadata_destination, malloc):
        raise NotImplementedError

    @abstractmethod
    def get_decoder_class(self) -> Type[Operation]:
        raise NotImplementedError



from .cutout import Cutout
from .flip import RandomHorizontalFlip
from .ops import ToTensor, ToDevice, ToTorchImage, Convert, View
from .common import Squeeze
from .random_resized_crop import RandomResizedCrop
from .poisoning import Poison
from .replace_label import ReplaceLabel
from .normalize import NormalizeImage
from .translate import RandomTranslate
from .mixup import ImageMixup, LabelMixup, MixupToOneHot
from .module import ModuleWrapper
from .color_jitter import RandomBrightness, RandomContrast, RandomSaturation

__all__ = ['ToTensor', 'ToDevice',
           'ToTorchImage', 'NormalizeImage',
           'Convert',  'Squeeze', 'View',
           'RandomResizedCrop', 'RandomHorizontalFlip', 'RandomTranslate',
           'Cutout', 'ImageMixup', 'LabelMixup', 'MixupToOneHot',
           'Poison', 'ReplaceLabel',
           'ModuleWrapper',
           'RandomBrightness', 'RandomContrast', 'RandomSaturation']



"""
Mixup augmentation for images and labels (https://arxiv.org/abs/1710.09412)
"""
from typing import Tuple

from numba import objmode
import numpy as np
import torch as ch
import torch.nn.functional as F
from dataclasses import replace
from typing import Callable, Optional, Tuple
from ..pipeline.allocation_query import AllocationQuery
from ..pipeline.operation import Operation
from ..pipeline.state import State
from ..pipeline.compiler import Compiler

class ImageMixup(Operation):
    """Mixup for images. Operates on raw arrays (not tensors).

    Parameters
    ----------
    alpha : float
        Mixup parameter alpha
    same_lambda : bool
        Whether to use the same value of lambda across the whole batch, or an
        individually sampled lambda per image in the batch
    """

    def __init__(self, alpha: float, same_lambda: bool):
        super().__init__()
        self.alpha = alpha
        self.same_lambda = same_lambda

    def generate_code(self) -> Callable:
        alpha = self.alpha
        same_lam = self.same_lambda
        my_range = Compiler.get_iterator()

        def mixer(images, dst, indices):
            np.random.seed(indices[-1])
            num_images = images.shape[0]
            lam = np.random.beta(alpha, alpha) if same_lam else \
                  np.random.beta(alpha, alpha, num_images)
            for ix in my_range(num_images):
                l = lam if same_lam else lam[ix]
                dst[ix] = l * images[ix] + (1 - l) * images[ix - 1]

            return dst

        mixer.is_parallel = True
        mixer.with_indices = True

        return mixer

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        return (previous_state, AllocationQuery(shape=previous_state.shape,
                                                dtype=previous_state.dtype))

class LabelMixup(Operation):
    """Mixup for labels. Should be initialized in exactly the same way as
    :cla:`ffcv.transforms.ImageMixup`.
    """
    def __init__(self, alpha: float, same_lambda: bool):
        super().__init__()
        self.alpha = alpha
        self.same_lambda = same_lambda

    def generate_code(self) -> Callable:
        alpha = self.alpha
        same_lam = self.same_lambda
        my_range = Compiler.get_iterator()

        def mixer(labels, temp_array, indices):
            num_labels = labels.shape[0]
            # permutation = np.random.permutation(num_labels)
            np.random.seed(indices[-1])
            lam = np.random.beta(alpha, alpha) if same_lam else \
                  np.random.beta(alpha, alpha, num_labels)

            for ix in my_range(num_labels):
                temp_array[ix, 0] = labels[ix][0]
                temp_array[ix, 1] = labels[ix - 1][0]
                temp_array[ix, 2] = lam if same_lam else lam[ix]

            return temp_array

        mixer.is_parallel = True
        mixer.with_indices = True

        return mixer

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        return (replace(previous_state, shape=(3,), dtype=np.float32),
                AllocationQuery((3,), dtype=np.float32))

class MixupToOneHot(Operation):
    def __init__(self, num_classes: int):
        super().__init__()
        self.num_classes = num_classes

    def generate_code(self) -> Callable:
        def one_hotter(mixedup_labels, dst):
            dst.zero_()
            N = mixedup_labels.shape[0]
            dst[ch.arange(N), mixedup_labels[:, 0].long()] = mixedup_labels[:, 2]
            mixedup_labels[:, 2] *= -1
            mixedup_labels[:, 2] += 1
            dst[ch.arange(N), mixedup_labels[:, 1].long()] = mixedup_labels[:, 2]
            return dst

        return one_hotter

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        # Should already be converted to tensor
        assert not previous_state.jit_mode
        return (replace(previous_state, shape=(self.num_classes,)), \
                AllocationQuery((self.num_classes,), dtype=previous_state.dtype, device=previous_state.device))


import numpy as np
from assertpy import assert_that
from numpy.random import shuffle
from torch.utils.data import Dataset
import logging
import os
from assertpy import assert_that
from tempfile import NamedTemporaryFile
from multiprocessing import cpu_count

from ffcv.writer import DatasetWriter
from ffcv.reader import Reader
from ffcv.fields import IntField, FloatField, BytesField

numba_logger = logging.getLogger('numba')
numba_logger.setLevel(logging.WARNING)

class DummyDataset(Dataset):

    def __init__(self, l):
        self.l = l

    def __len__(self):
        return self.l

    def __getitem__(self, index):
        if index > self.l:
            raise IndexError()
        return (index, np.sin(index))

class DummyDatasetWithData(Dataset):

    def __init__(self, l):
        self.l = l

    def __len__(self):
        return self.l

    def __getitem__(self, index):
        if index > self.l:
            raise IndexError()
        return (index, np.zeros(2))

def validate_simple_dataset(name, length, shuffled=False):
    reader = Reader(name)
    assert_that(reader.handlers).is_length(2)
    assert_that(reader.handlers['index']).is_instance_of(IntField)
    assert_that(reader.handlers['value']).is_instance_of(FloatField)
    assert_that(reader.alloc_table).is_length(0)
    assert_that(reader.metadata).is_length(length)
    if shuffled:
        assert_that((reader.metadata['f0'] == np.arange(length).astype('int')).all()).is_false()
        assert_that(set(reader.metadata['f0'])).is_equal_to(set(np.arange(length).astype('int')))
    else:
        assert_that((reader.metadata['f0'] == np.arange(length).astype('int')).all()).is_true()
    assert_that((np.sin(reader.metadata['f0']) == reader.metadata['f1']).all()).is_true()

def test_write_shuffle():
    length = 600
    with NamedTemporaryFile() as handle:
        name = handle.name
        dataset = DummyDataset(length)

        writer = DatasetWriter(name, {
            'index': IntField(),
            'value': FloatField()
        })

        writer.from_indexed_dataset(dataset, shuffle_indices=True)

        validate_simple_dataset(name, length, shuffled=True)

def test_write_simple():
    length = 600
    with NamedTemporaryFile() as handle:
        name = handle.name
        dataset = DummyDataset(length)
        writer = DatasetWriter(name, {
            'index': IntField(),
            'value': FloatField()
        })

        writer.from_indexed_dataset(dataset)

        validate_simple_dataset(name, length)

def test_multiple_workers():
    length = 600
    with NamedTemporaryFile() as handle:
        name = handle.name
        dataset = DummyDataset(length)
        writer = DatasetWriter(name, {
            'index': IntField(),
            'value': FloatField()
        }, num_workers=min(30, cpu_count()))

        writer.from_indexed_dataset(dataset, chunksize=10000)

        validate_simple_dataset(name, length)


def test_super_long():
    length = 600000
    with NamedTemporaryFile() as handle:
        name = handle.name
        dataset = DummyDataset(length)
        writer = DatasetWriter(name, {
            'index': IntField(),
            'value': FloatField()
        }, num_workers=min(30, cpu_count()))

        writer.from_indexed_dataset(dataset, chunksize=10000)

        validate_simple_dataset(name, length)

def test_small_chunks_multiple_workers():
    length = 600
    with NamedTemporaryFile() as handle:
        name = handle.name
        dataset = DummyDatasetWithData(length)
        writer = DatasetWriter(name, {
            'index': IntField(),
            'value': BytesField()
        }, num_workers=min(30, cpu_count()))

        writer.from_indexed_dataset(dataset, chunksize=1)